[inherit('SYS$LIBRARY:STARLET.PEN', 
	 'SYS$LIBRARY:PASCAL$LIB_ROUTINES.PEN', 
	 'SYS$LIBRARY:PASCAL$STR_ROUTINES.PEN'
	)]
PROGRAM HTSHELF(input, output);
(*
 * A program to traverse bookshelf files
 * It is intended for use with the OSU HTTP_SERVER
 * Author: Stellan Lagerstrm <stellanl@e.kth.se>
 * Copyright: None. Share and, if possible, enjoy
 *
 * Book references from here will be URLs like
 *  /book/ABCNKVD.DECW$BOOK	! the book as a whole: output the TOC
 * not the more specific
 *  /book/ABCNKVD.DECW$BOOK/7	! a page in the book
 *
 * Version
 *  0.19	6-JAN-1996 02:46	Fixed escaping URLs
 *)

CONST	(* adapt these to your local environment *)
  book_icon   = '/book_figures/book.gif';
  shelf_icon  = '/book_figures/shelf.gif';
  book_url    = '/book/';
  defaultname = 'DECW$BOOK:LIBRARY.DECW$BOOKSHELF;*';

CONST 
  crlf	     = CHR(13)+CHR(10);
  maxparts   = 20;

TYPE
  urlstr = VARYING [1024] OF CHAR;		(* may be too little... *)
  str    = PACKED ARRAY [1..80] OF CHAR;
  strp   = ^str;
  $UWORD = [word]0..65535;

VAR
  url		: ARRAY[1..maxparts]OF STRING(255);
  url_parts	: INTEGER VALUE 0;
  log		: TEXT;
  shelf		: [volatile]STRING(40);
  param		: ARRAY[1..8]OF STRING(255);
  default_title	: STRING(255);

  done_title    : BOOLEAN VALUE FALSE;
  done_header   : BOOLEAN VALUE FALSE;

  rec : [volatile]VARYING[1024] OF CHAR;	(* we hope *)

  result : [volatile]STRING(255);
  relate : [volatile]STRING(255);
  expand : [volatile]STRING(255);

  done_shelves:STRING(8190) VALUE '';	(* sigh... *)

  fab:FAB$TYPE VALUE [fab$b_bid:fab$c_bid; fab$b_bln:fab$c_bln;
		      fab$v_nam:TRUE;	(* to open from NAM block *)
		      fab$v_sqo:TRUE;	(* performance *)
		      fab$v_get:TRUE;
		      fab$v_shrget:TRUE;
		      OTHERWISE ZERO];
  nam:NAM$TYPE VALUE [nam$b_bid:nam$c_bid; nam$b_bln:nam$c_bln; OTHERWISE ZERO];

  b:STRING(255) VALUE '';

  [ASYNCHRONOUS,external(str$upcase)]
    FUNCTION str_upcase (VAR destination_string	: [volatile]STRING;
			     source_string	: [volatile]STRING):INTEGER;EXTERNAL;


  PROCEDURE WR(VAR s:[readonly]STRING);
  BEGIN
    IF LENGTH(b)+LENGTH(s)>b.capacity THEN BEGIN WRITELN(b); b:=''; END;
    b:=b+s;
  END;


  PROCEDURE FLUSH(i:INTEGER:=MAXINT);
  BEGIN				(* make sure of i chars free in buffer *)
    IF b.capacity-LENGTH(b)<i THEN BEGIN WRITELN(b); b:=''; END;
  END;


  FUNCTION TRNLOG(l:STRING):urlstr;
  TYPE
    item3 = PACKED RECORD len,item:$UWORD; adr:INTEGER; retlenadr:^$UWORD; END ;
  VAR
    t:[volatile]urlstr;
    itm:PACKED ARRAY[0..1]OF item3 VALUE ZERO;
  BEGIN
    itm[0].item:=LNM$_STRING;
    itm[0].len:=SIZE(t.body);
    itm[0].adr:=IADDRESS(t.body);
    itm[0].retlenadr:=ADDRESS(t.length);
    IF NOT ODD($TRNLNM(,'LNM$FILE_DEV',l,,itm))
      THEN t:='';
    TRNLOG:=t;
  END;


  FUNCTION MSG(i:[unsafe]INTEGER):urlstr;
  VAR t:urlstr;
  BEGIN
    $GETMSG(i,t.length,t.body,15);
    MSG:=t; (* '%x'+HEX(i,8);		(* NYI *)
  END;


  PROCEDURE BUG(s:STRING);
  BEGIN
    WR('<PRE>'); WR(s); WR('</PRE>'); WR(crlf);
  END;

  PROCEDURE ESCAPE(s:STRING); (* escape characters in an URL *)
  VAR I:INTEGER;
  BEGIN
    FOR i:=1 TO LENGTH(s)
      DO CASE s[i] OF
	'a'..'z', 
	'A'..'Z', 
	'0'..'9', 
	'_', 
	'-', 
	'.': WR(s[i]);
	OTHERWISE WR('%'+HEX(s[i], 2));
      END;
  END;

  PROCEDURE HTML_ESCAPE(s:STRING);
  VAR I:INTEGER;
  BEGIN
    FOR i:=1 TO LENGTH(s)
      DO CASE s[i] OF
	'<': WR('&lt;');
	'>': WR('&gt;');
	'&': WR('&amp;');
	OTHERWISE
	     WR(s[i]);
      END;
  END;

  PROCEDURE UNESCAPE(s:STRING; VAR t:STRING);
  VAR
    i : INTEGER;
    a : ARRAY [CHAR] OF INTEGER VALUE ['1':1;'2':2;'3':3;'4':4;'5':5;
				       '6':6;'7':7;'8':8;'9':9;
				       'A','a' : 10;
				       'B','b' : 11;
				       'C','c' : 12;
				       'D','d' : 13;
				       'E','e' : 14;
				       'F','f' : 15;
				       OTHERWISE 0];
  BEGIN
    t:='';
    i:=1;
    WHILE (i <= LENGTH(s)) DO BEGIN
      CASE s[i] OF
	'%': IF i < LENGTH(s)-2 THEN BEGIN
	  t:=t+CHR(a[s[i+1]]*16+a[s[i+2]]);
	  i:=i+2;
	END;
	OTHERWISE t:=t+s[i];
      END;
      i:=i+1;
    END;
  END;

  PROCEDURE DFLT(f:STRING; d:STRING; t:STRING);	(* syntax-only defaulting *)
  TYPE
    item2 = PACKED RECORD len,item:$UWORD; adr:strp END ;
  VAR
    ff,df:[long,volatile]FLDFLAGS$TYPE;
    fi,di:PACKED ARRAY[0..6]OF item2 VALUE [
            0:[item:FSCN$_NODE;OTHERWISE ZERO];
            1:[item:FSCN$_DEVICE;OTHERWISE ZERO];
            2:[item:FSCN$_DIRECTORY;OTHERWISE ZERO];
            3:[item:FSCN$_NAME;OTHERWISE ZERO];
            4:[item:FSCN$_TYPE;OTHERWISE ZERO];
            5:[item:FSCN$_VERSION;OTHERWISE ZERO]; (* weird people, but... *)
            6:[OTHERWISE ZERO]
	    ];
  BEGIN
    IF  ODD($FILESCAN(f,fi,ff::UNSIGNED))
    AND ODD($FILESCAN(d,di,df::UNSIGNED)) THEN BEGIN

      IF ff.FSCN$V_NODE
        THEN ESCAPE(SUBSTR(fi[0].adr^,1,fi[0].len))
	ELSE IF df.FSCN$V_NODE
	       THEN ESCAPE(SUBSTR(di[0].adr^,1,di[0].len));

      IF ff.FSCN$V_DEVICE
        THEN WR(SUBSTR(fi[1].adr^,1,fi[1].len))
	ELSE IF df.FSCN$V_DEVICE
	       THEN WR(SUBSTR(di[1].adr^,1,di[1].len));

      IF ff.FSCN$V_DIRECTORY
        THEN ESCAPE(SUBSTR(fi[2].adr^,1,fi[2].len)) (* hide brackets?? *)
	ELSE IF df.FSCN$V_DIRECTORY
	       THEN ESCAPE(SUBSTR(di[2].adr^,1,di[2].len)); (* -"- *)

      IF ff.FSCN$V_NAME			(* This SHOULD be present! *)
        THEN WR(SUBSTR(fi[3].adr^,1,fi[3].len))
	ELSE WR('NONAME');

      IF ff.FSCN$V_TYPE
        THEN WR(SUBSTR(fi[4].adr^,1,fi[4].len))
	ELSE WR(t);

      IF ff.FSCN$V_VERSION
        THEN WR(SUBSTR(fi[5].adr^,1,fi[5].len));
    END;
  END;

  PROCEDURE DO_SHELF_LINE;
  VAR
    what  : VARYING[40]OF CHAR;
    where : VARYING[255]OF CHAR;
    title : VARYING[255]OF CHAR;
    i	  : INTEGER;
  BEGIN
    IF NOT EQ(rec, '') THEN BEGIN

      IF (rec[1]='!') THEN BEGIN (* comment *)
	IF (LENGTH(rec)>5) AND_THEN EQ(SUBSTR(rec,2,5),'HTML\') (* embedded HTML *)
	  THEN WR(SUBSTR(rec,7,LENGTH(rec)-6)+crlf);
      END ELSE

      IF  ODD(STR$ELEMENT(%DESCR what,  0, '\', rec)) (* not comment *)
      AND ODD(STR$ELEMENT(%DESCR where, 1, '\', rec))
      AND ODD(STR$ELEMENT(%DESCR title, 2, '\', rec)) THEN BEGIN

	STR$UPCASE(%DESCR what, what);

	IF what='SHELF' THEN BEGIN (* another mission for ourselves *)
	  IF NOT done_title
	    THEN WR('<HEAD><TITLE>'+ default_title+ '</TITLE></HEAD><BODY>');
	  done_title:=TRUE;
	  IF NOT done_header
	    THEN WR('<H2>'+default_title+'</H2>');
	  done_header:=TRUE;
	  WR('<IMG SRC="'+ shelf_icon+ '">');
	  FLUSH(80+LENGTH(param[2])+LENGTH(title)+LENGTH(where));
	  WR(' <A HREF="/');
	  FOR i:=1 TO url_parts DO BEGIN ESCAPE(url[i]); WR('/'); END;
	  ESCAPE(title);	WR('?');
	  DFLT(where,shelf,'.DECW$BOOKSHELF'); 	WR('">');
	  HTML_ESCAPE(title);	WR('</A><BR>'+crlf);
	END ELSE

	IF what='BOOK' THEN BEGIN   (* send in the clowns *)   
	  IF NOT done_title
	    THEN WR('<HEAD><TITLE>'+ default_title+ '</TITLE></HEAD><BODY>');
	  done_title:=TRUE;
	  IF NOT done_header
	    THEN WR('<H2>'+default_title+'</H2>');
	  done_header:=TRUE;
	  IF NOT done_title
	    THEN WR('<HEAD><TITLE>' + default_title + '</TITLE></HEAD><BODY>' +
	            '<H2>' + default_title + '</H2>');
	  done_title:=TRUE;
	  WR('<IMG SRC="'+book_icon+'">');
	  FLUSH(80+LENGTH(where));
	  WR(' <A HREF="'+book_url);
	  DFLT(where,shelf,'.DECW$BOOK'); 	WR('">');
	  HTML_ESCAPE(title);	WR('</A><BR>'+crlf);
	END ELSE

	IF what='TITLE' THEN BEGIN
	  IF NOT done_title
	    THEN WR('<HEAD><TITLE>'+ title+ '</TITLE></HEAD><BODY>');
	  done_title:=TRUE;
	  IF NOT done_header
	    THEN WR('<H2>'+title+'</H2>');
	  done_header:=TRUE;
	END ELSE
	  BUG('%HTSHELF-E-SYNTAX, error in shelf '+shelf+crlf+' \'+rec+'\');

      END ELSE
        BUG('%HTSHELF-E-SYNTAX, error in shelf '+shelf+crlf+' \'+rec+'\');
    END;
  END;

  PROCEDURE DO_SHELF_FILE;
  VAR
    rab:RAB$TYPE VALUE [rab$b_bid:rab$c_bid; rab$b_bln:rab$c_bln;
(*		        rab$v_rah:TRUE;*)
			OTHERWISE ZERO];
    name:STRING(255);
    
  BEGIN
    name:=SUBSTR(nam.nam$l_dev::strp^,1,nam.nam$b_dev)
         +SUBSTR(nam.nam$l_dir::strp^,1,nam.nam$b_dir)
	 +SUBSTR(nam.nam$l_name::strp^,1,nam.nam$b_name)
	 +SUBSTR(nam.nam$l_type::strp^,1,nam.nam$b_type);
    WR('<!-- %HTSHELF-I-LIBNAME, '+ name+'-->'+ crlf);
    IF INDEX(done_shelves,name+'') = 0 THEN BEGIN  (* if not had it before *)

      done_shelves:=done_shelves+name+'';
      IF url_parts>2			(* did we get a title in the url? *)	     
        THEN default_title:=url[url_parts]
	ELSE default_title:=name;

      rab.rab$l_fab := IADDRESS(fab);
      rab.rab$l_ubf := IADDRESS(rec.body);
      rab.rab$w_usz := SIZE(rec.body);
      IF ODD($OPEN(fab)) THEN BEGIN
	IF ODD($CONNECT(rab)) THEN BEGIN

	  WHILE ODD($GET(rab)) DO BEGIN
	    rec.length:=rab.rab$w_rsz;
	    DO_SHELF_LINE;
	  END;
	  IF rab.rab$l_sts<>RMS$_EOF
	    THEN WR('%HTSHELF-E-GET, sts '+MSG(rab.rab$l_sts)+ crlf);

	  $DISCONNECT(rab);

	END ELSE
	  WR('<PRE>%HTSHELF-F-CONNECT, sts '+ MSG(rab.rab$l_sts)+ '</PRE>'+crlf);
	$CLOSE(fab);
      END ELSE
	WR('<PRE>%HTSHELF-F-OPEN, sts '+ MSG(fab.fab$l_sts)+ '</PRE>'+crlf);
    END;
  END;

  PROCEDURE GET_PARAMS;
  VAR
    par	  : STRING(255);
    j	  : INTEGER;
    p	  : INTEGER VALUE 0;
    c	  : CHAR VALUE ' ';
    q	  : BOOLEAN VALUE FALSE;
    myurl : STRING(1024);		(* the url of this shelf *)
  BEGIN
    (* get the search argument, which is used as the shelf id *)
    WRITELN('<DNETARG>');
    READLN(shelf);
    UNESCAPE(shelf, shelf);

    (* We were activated as a foreign command *)
    (* with  P1= 'GET' P2=url P3=HTTP-proto *)
    (* URL should look like /htbindir/bookshelf/a/b/c ?shelf *)
    LIB$GET_FOREIGN(par.body, , par.length);
    FOR j:=1 TO LENGTH(par) DO BEGIN
      CASE par[j] OF
	'"': BEGIN IF NOT q AND (c=' ') THEN p:=p+1; q:=NOT q; END;
	' ': IF q THEN param[p]:=param[p]+' ';
	OTHERWISE
	  IF NOT q AND (c=' ') THEN p:=p+1;
	  param[p]:=param[p]+par[j];
      END;
      c:=par[j];
    END;
    WRITELN(log, '%HTSHELF-I-PARAMS, ', par); (* debugging *)
    UNESCAPE(param[2], myurl);
    FOR j:=1 TO LENGTH(myurl) DO CASE myurl[j] OF
      '/': url_parts:=MIN(maxparts,url_parts+1);
      OTHERWISE	
           url[url_parts]:=url[url_parts]+myurl[j];
    END;
    
    WRITELN(log, '%HTSHELF-I-URL, ', myurl); (* debugging *)
    
  END;


BEGIN
  fab.fab$l_nam:=IADDRESS(nam);

  nam.nam$l_esa:=IADDRESS(expand.body);
  nam.nam$b_ess:=SIZE(expand.body);
  nam.nam$l_rsa:=IADDRESS(result.body);
  nam.nam$b_rss:=SIZE(result.body);
  nam.nam$l_rlf:=0; (*another NAM block! *)
  nam.nam$b_rsl:=0;

  OPEN(output, 'NET_LINK:', OLD);	REWRITE(output);
  OPEN(input, 'NET_LINK:', OLD);	RESET(input);
  OPEN(log, 'SYS$OUTPUT:', UNKNOWN);	REWRITE(log);

  GET_PARAMS;

  WRITELN('<DNETCGI>');
  WRITELN('Content-Type: text/html', crlf, crlf);
  WR('<HTML>'+crlf);
  (*WR('<!-- %HTSHELF-I-URL, '+ url+ ' ? '+shelf+'-->'); (* debugging *)
  IF NOT EQ(shelf, '') AND_THEN (shelf[1]='?')
    THEN shelf:=SUBSTR(shelf, 2, LENGTH(shelf)-1);
  IF EQ(shelf, '')
    THEN shelf:=TRNLOG("DECW$BOOKSHELF");

  STR_UPCASE(shelf, shelf);


  WRITELN(log, '%HTSHELF-I-SHELF, ', shelf); (* debugging *)
  fab.fab$l_fna:=IADDRESS(shelf.body);
  fab.fab$b_fns:=LENGTH(shelf);
  fab.fab$l_dna:=IADDRESS(defaultname);	      (* is a constant! *)	      
  fab.fab$b_dns:=LENGTH(defaultname);

  IF ODD($PARSE(fab)) THEN BEGIN
    expand.length:=nam.nam$b_esl;
    WR('<!-- %HTSHELF-I-PARSE, '+ expand+ '-->'+ crlf); (* debugging *)
    
    WHILE ODD($SEARCH(fab)) DO BEGIN
      result.length:=nam.nam$b_rsl;
      WR('<!--  %HTSHELF-I-SEARCH, '+ result+ '-->'+crlf);
      DO_SHELF_FILE;
      IF NOT EQ(shelf,'') THEN EXIT;
      done_header:=FALSE;
    END;
  END ELSE
    BUG('%HTSHELF-F-PARSE, sts '+MSG(fab.fab$l_sts)); (* debugging *)
  WR('</HTML>'+crlf);
  FLUSH;

  WRITELN('</DNETCGI>');	(* bye, bye *)
  CLOSE(output);
END.


